# ruff: noqa

"""init."""
from .bijankhan_reader import BijankhanReader
from .dadegan_reader import DadeganReader
from .universal_dadegan_reader import UniversalDadeganReader
from .degarbayan_reader import DegarbayanReader
from .hamshahri_reader import HamshahriReader
from .mirastext_reader import MirasTextReader
from .persica_reader import PersicaReader
from .peykare_reader import PeykareReader
from .quran_reader import QuranReader
from .sentipers_reader import SentiPersReader
from .tnews_reader import TNewsReader
from .treebank_reader import TreebankReader
from .verbvalency_reader import VerbValencyReader
from .wikipedia_reader import WikipediaReader
